@@ -76,6 +76,14 @@ w_callback(picoev_loop* loop, int fd, int events, void* cb_arg);
 static inline void
 resume_wsgi_app(ClientObject *pyclient, picoev_loop* loop);
 
+static inline void
+prepare_call_wsgi(client_t *client);
+
+static inline void
+call_wsgi_app(client_t *client, picoev_loop* loop);
+
+static inline int
+check_status_code(client_t *client);
 
 static inline int
 setsig(int sig, void* handler)
@@ -145,16 +153,12 @@ new_client_t(int client_fd, char *remote_addr, uint32_t remote_port){
     client = alloc_client_t();
     //client = PyMem_Malloc(sizeof(client_t)); 
     //memset(client, 0, sizeof(client_t));
-    
-    //printf("size %d\n", sizeof(client_t));
 
     client->fd = client_fd;
-    
+    client->request_queue = new_request_queue();    
     client->remote_addr = remote_addr;
     client->remote_port = remote_port;
-    client->req = new_request();    
     client->body_type = BODY_TYPE_NONE;
-    //printf("input_buf_size %d\n", client->input_buf_size);
     return client;
 }
 
@@ -174,23 +178,25 @@ clean_cli(client_t *client)
 #ifdef DEBUG
     printf("clean_cli environ status_code %d address %p \n", client->status_code, client->environ);
 #endif
-    
-    // force clear
-    PyDict_Clear(client->environ);
-    
-    Py_CLEAR(client->environ);
-    if(client->body_type == BODY_TYPE_TMPFILE){
-        if(client->body){
-            fclose(client->body);
+    if(client->environ){ 
+        PyDict_Clear(client->environ);
+        Py_DECREF(client->environ);
+    }
+    if(client->body){
+        if(client->body_type == BODY_TYPE_TMPFILE){
+            if(client->body){
+                fclose(client->body);
+            }
+        }else{
+            Py_CLEAR(client->body);
         }
-    }else{
-        Py_CLEAR(client->body);
     }
-    if(client->http != NULL){
-        PyMem_Free(client->http);
-        client->http = NULL;
-    }
-    
+    client->header_done = 0;
+    client->response_closed = 0;
+    client->chunked_response = 0;
+    client->content_length_set = 0;
+    client->content_length = 0;
+    client->write_bytes = 0;
 }
 
 static inline void 
@@ -202,11 +208,28 @@ close_conn(client_t *cli, picoev_loop* loop)
     }
 
     picoev_del(loop, cli->fd);
+    clean_cli(cli);
 
 #ifdef DEBUG
+    printf("start close client:%p fd:%d status_code %d \n", cli, cli->fd, cli->status_code);
     printf("picoev_del client:%p fd:%d \n", cli, cli->fd);
+    printf("remain http pipeline size :%d \n", cli->request_queue->size);
 #endif
-    clean_cli(cli);
+    
+    if(cli->request_queue->size > 0){
+        if(check_status_code(cli) > 0){
+            //process pipeline 
+            prepare_call_wsgi(cli);
+            call_wsgi_app(cli, loop);
+        }
+        return ;
+    }
+
+    if(cli->http != NULL){
+        PyMem_Free(cli->http);
+    }
+
+    free_request_queue(cli->request_queue);
     if(!cli->keep_alive){
         close(cli->fd);
 #ifdef DEBUG
@@ -221,9 +244,53 @@ close_conn(client_t *cli, picoev_loop* loop)
     }
     //PyMem_Free(cli);
     dealloc_client(cli);
+#ifdef DEBUG
+    printf("********************\n\n");
+#endif
 
 }
 
+static inline void
+set_current_request(client_t *client)
+{
+    request *req;
+    req = shift_request(client->request_queue);
+    client->bad_request_code = req->bad_request_code;
+    client->body = req->body;
+    client->body_type = req->body_type;
+    client->environ = req->env;
+    free_request(req);
+    client->req = NULL;
+}
+
+static inline void
+set_bad_request_code(client_t *client, int status_code)
+{
+    request *req;
+    req = client->request_queue->tail;
+    req->bad_request_code = status_code;
+}
+
+static inline int
+check_status_code(client_t *client)
+{
+    request *req;
+    req = client->request_queue->head;
+    if(req->bad_request_code > 200){
+        //error
+        //shift 
+#ifdef DEBUG
+        printf("bad_request_code \n");
+#endif 
+        set_current_request(client);
+        send_error_page(client);
+        close_conn(client, main_loop);
+        return 0;
+    }
+    return 1;
+    
+}
+
 static inline int
 process_resume_wsgi_app(ClientObject *pyclient)
 {
@@ -292,6 +359,7 @@ process_wsgi_app(client_t *cli)
     pyclient = (ClientObject *)current_client;
 
 #ifdef DEBUG
+    printf("start client %p \n", cli);
     printf("start environ %p \n", cli->environ);
 #endif
 
@@ -548,7 +616,35 @@ prepare_call_wsgi(client_t *client)
 {
     PyObject *input = NULL, *object = NULL, *c = NULL;
     char *val;
-
+    
+    set_current_request(client);
+    
+    //check Expect
+    if(client->http->http_minor == 1){
+        c = PyDict_GetItemString(client->environ, "HTTP_EXPECT");
+        if(c){
+            val = PyString_AS_STRING(c);
+            if(!strcasecmp(val, "100-continue")){
+                int ret = write(client->fd, "HTTP/1.1 100 Continue\r\n\r\n", 25);
+                if(ret < 0){
+                    PyErr_SetFromErrno(PyExc_IOError);
+                    write_error_log(__FILE__, __LINE__); 
+                    client->keep_alive = 0;
+                    client->status_code = 500;
+                    send_error_page(client);
+                    close_conn(client, main_loop);
+                    return;
+                }
+            }else{
+                //417
+                client->keep_alive = 0;
+                client->status_code = 417;
+                send_error_page(client);
+                close_conn(client, main_loop);
+                return;
+            }
+        }
+    }
     if(client->body_type == BODY_TYPE_TMPFILE){
         FILE *tmp = (FILE *)client->body;
         fflush(tmp);
@@ -577,18 +673,34 @@ prepare_call_wsgi(client_t *client)
         //Py_DECREF(object);
         Py_DECREF(input);
     }
+
     if(is_keep_alive){
         //support keep-alive
         c = PyDict_GetItemString(client->environ, "HTTP_CONNECTION");
-        if(c){
-            val = PyString_AS_STRING(c);
-            if(!strcasecmp(val, "keep-alive")){
+        if(client->http->http_minor == 1){
+            //HTTP 1.1
+            if(c){
+                val = PyString_AS_STRING(c);
+                if(!strcasecmp(val, "close")){
+                    client->keep_alive = 0;
+                }else{
+                    client->keep_alive = 1;
+                }
+            }else{
                 client->keep_alive = 1;
+            }
+        }else{
+            //HTTP 1.0
+            if(c){
+                val = PyString_AS_STRING(c);
+                if(!strcasecmp(val, "keep-alive")){
+                    client->keep_alive = 1;
+                }else{
+                    client->keep_alive = 0;
+                }
             }else{
                 client->keep_alive = 0;
             }
-        }else{
-            client->keep_alive = 0;
         }
     }
 }
@@ -599,6 +711,7 @@ r_callback(picoev_loop* loop, int fd, int events, void* cb_arg)
     client_t *cli = ( client_t *)(cb_arg);
     PyObject *body = NULL;
     char *key = NULL;
+    int finish = 0, nread;
 
     if ((events & PICOEV_TIMEOUT) != 0) {
 
@@ -607,15 +720,15 @@ r_callback(picoev_loop* loop, int fd, int events, void* cb_arg)
 #endif
         //timeout
         cli->keep_alive = 0;
-        close_conn(cli, loop);
+        if(cli->request_queue->size > 0){
+            //piplining
+            set_bad_request_code(cli, 408);
+            finish = 1;
+        }else{
+            close_conn(cli, loop);
+        }
     
     } else if ((events & PICOEV_READ) != 0) {
-
-#ifdef DEBUG
-        printf("ready read \n");
-#endif
-        /* update timeout, and read */
-        int finish = 0, nread;
         char buf[INPUT_BUF_SIZE];
         ssize_t r;
         if(!cli->keep_alive){
@@ -626,63 +739,81 @@ r_callback(picoev_loop* loop, int fd, int events, void* cb_arg)
             case 0: 
                 cli->keep_alive = 0;
                 //503??
-                cli->status_code = 503;
-                send_error_page(cli);
-                close_conn(cli, loop);
-                return;
+                if(cli->request_queue->size > 0){
+                    //piplining
+                    set_bad_request_code(cli, 503);
+                    finish = 1;
+                }else{
+                    cli->status_code = 503;
+                    send_error_page(cli);
+                    close_conn(cli, loop);
+                    return;
+                }
             case -1: /* error */
                 if (errno == EAGAIN || errno == EWOULDBLOCK) { /* try again later */
                     break;
                 } else { /* fatal error */
-                    if(cli->keep_alive && errno == ECONNRESET){
-                        
-                        cli->keep_alive = 0;
-                        cli->status_code = 500;
-                        cli->header_done = 1;
-                        cli->response_closed = 1;
-                        
-                    }else{
-                        PyErr_SetFromErrno(PyExc_IOError);
-                        write_error_log(__FILE__, __LINE__); 
-                        cli->keep_alive = 0;
-                        cli->status_code = 500;
+                    if(cli->request_queue->size > 0){
+                        //piplining
+                        set_bad_request_code(cli, 500);
                         if(errno != ECONNRESET){
-                            send_error_page(cli);
-                        }else{
+                            PyErr_SetFromErrno(PyExc_IOError);
+                            write_error_log(__FILE__, __LINE__); 
+                        }
+                        finish = 1;
+                    }else{
+                        if(cli->keep_alive && errno == ECONNRESET){
+                        
+                            cli->keep_alive = 0;
+                            cli->status_code = 500;
                             cli->header_done = 1;
                             cli->response_closed = 1;
+                        
+                        }else{
+                            PyErr_SetFromErrno(PyExc_IOError);
+                            write_error_log(__FILE__, __LINE__); 
+                            cli->keep_alive = 0;
+                            cli->status_code = 500;
+                            if(errno != ECONNRESET){
+                                send_error_page(cli);
+                            }else{
+                                cli->header_done = 1;
+                                cli->response_closed = 1;
+                            }
                         }
+                        close_conn(cli, loop);
+                        return;
                     }
-                    close_conn(cli, loop);
-                    return;
                 }
                 break;
             default:
-
+#ifdef DEBUG
+                printf("********************\n%s\n", buf);
+#endif
                 nread = execute_parse(cli, buf, r);
 #ifdef DEBUG
                 printf("read request fd %d readed %d nread %d \n", cli->fd, r, nread);
-                printf("%s\n", buf);
 #endif
                 
                 if(cli->bad_request_code > 0){
 #ifdef DEBUG
                     printf("fd %d bad_request code %d \n",cli->fd,  cli->bad_request_code);
 #endif
-                    send_error_page(cli);
-                    close_conn(cli, loop);
-                    return;
+                    set_bad_request_code(cli, cli->bad_request_code);
+                    ///force end
+                    finish = 1;
+                    break;
                 }
 
                 if(!cli->upgrade && nread != r){
                     // parse error
 #ifdef DEBUG
-                    printf("fd %d parse error %d \n", cli->fd, cli->bad_request_code);
+                    printf("fd %d parse error Bad Request %d \n", cli->fd, cli->bad_request_code);
 #endif
-                    cli->bad_request_code = 400;
-                    send_error_page(cli);
-                    close_conn(cli, loop);
-                    return;
+                    set_bad_request_code(cli, 400);
+                    ///force end
+                    finish = 1;
+                    break;
                 }
                 
 #ifdef DEBUG
@@ -694,7 +825,7 @@ r_callback(picoev_loop* loop, int fd, int events, void* cb_arg)
                         //WebSocket Key
                         key = buf + nread + 1;
                         body = PyString_FromStringAndSize(key, r - nread);
-                        cli->body = body;
+                        cli->request_queue->tail->body = body;
 
                     }
                     finish = 1;
@@ -702,12 +833,15 @@ r_callback(picoev_loop* loop, int fd, int events, void* cb_arg)
                 break;
         }
 
-        if(finish == 1){
-            picoev_del(loop, cli->fd);
+    }
+    if(finish == 1){
+        picoev_del(loop, cli->fd);
+        if(check_status_code(cli) > 0){
+            //current request ok
             prepare_call_wsgi(cli);
             call_wsgi_app(cli, loop);
-            return;
         }
+        return;
     }
 }
 
