@@ -1,15 +1,18 @@
 import collections
 import string
 import struct
+import socket
 
 try:
     from hashlib import md5
 except ImportError: #pragma NO COVER
     from md5 import md5
 
-from meinheld import server
+from meinheld import server, patch
 from meinheld.common import Continuation, CLIENT_KEY, CONTINUATION_KEY
-import greenlet
+patch.patch_socket()
+
+import socket
 
 class WebSocketMiddleware(object):
 
@@ -43,7 +46,7 @@ class WebSocketMiddleware(object):
 
         # Get the underlying socket and wrap a WebSocket class around it
         client = environ[CLIENT_KEY]
-        sock = server._get_socket_fromfd(client.get_fd())
+        sock = socket.fromfd(client.get_fd(), socket.AF_INET, socket.SOCK_STREAM)
         ws = WebSocket(sock, environ, protocol_version)
         
         # If it's new-version, we need to work out our challenge response
@@ -102,18 +105,15 @@ class WebSocketMiddleware(object):
             if result and response != -1:
                 ws = environ.pop('wsgi.websocket')
                 ws._send_closing_frame(True)
+                client = environ[CLIENT_KEY]
+                client.set_closed(1)
 
     def __call__(self, environ, start_response):
         client = environ[CLIENT_KEY]
-        g = client.get_greenlet()
-        if not g:
-            # new greenlet
-            g = greenlet.greenlet(self.spawn_call)
-            client.set_greenlet(g)
-            c = Continuation(client)
-            environ[CONTINUATION_KEY] = c
+        c = Continuation(client)
+        environ[CONTINUATION_KEY] = c
 
-        return g.switch(environ, start_response)
+        return self.spawn_call(environ, start_response)
 
 class WebSocketWSGI(object):
 
@@ -140,7 +140,8 @@ class WebSocketWSGI(object):
 
         # Get the underlying socket and wrap a WebSocket class around it
         client = environ[CLIENT_KEY]
-        sock = server._get_socket_fromfd(client.get_fd(), client)
+        sock = server._get_socket_fromfd(client.get_fd(), socket.AF_INET,
+                socket.SOCK_STREAM)
         ws = WebSocket(sock, environ, self.protocol_version)
         
         # If it's new-version, we need to work out our challenge response
@@ -332,6 +333,6 @@ class WebSocket(object):
         """Forcibly close the websocket; generally it is preferable to
         return from the handler method."""
         self._send_closing_frame()
-        #self.socket.shutdown(True)
+        self.socket.shutdown(True)
         self.socket.close()
 
